#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/cdrom.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include "Jcd_Drive.h"

/**
  Jcd - Java CD Audio Player
  Copyright (c) 1996.   Michael Hamilton (michael@actrix.gen.nz).
  All rights reserved.  See the README for full details
 */

/*
 * This code could be far more integrated into the Java code, but
 * I may want to used it with python (my prefered langauge), so
 * I'm avoiding this for now.  Ignoring supporting multi-disc players
 * for now, but the java class allows for it in the future.
 */

#ifdef TRUE
#undef TRUE
#undef FALSE
#endif

#define TRUE  1
#define FALSE 0

/*
 * The following are normally be by "javah Jcd.Drive" using Sun's JDK.
 * However kaffeh doesn't produce them - so we'll have to it ourselves.
 */
#ifndef Jcd_Drive_FRAMES_PER_SECOND
#  define Jcd_Drive_FRAMES_PER_SECOND 75L

#  define Jcd_Drive_STATUS_INVALID 0L
#  define Jcd_Drive_STATUS_PLAY 17L
#  define Jcd_Drive_STATUS_PAUSED 18L
#  define Jcd_Drive_STATUS_COMPLETED 19L
#  define Jcd_Drive_STATUS_ERROR 20L
#  define Jcd_Drive_STATUS_NOSTATUS 21L

#  define Jcd_Drive_FLAG_NONE 0L
#  define Jcd_Drive_FLAG_STOP_PLAY 1L
#endif

/*
 * Javah and Kaffeh differ over the type of long
 */
#ifdef Jcd_Drive_FRAMES_PER_SECOND
#  define Java_Int long		/* javah */
#else
#  define Java_Int jint		/* kaffeh */
#endif


#define FRAMES_PER_SECOND Jcd_Drive_FRAMES_PER_SECOND		
#define MAX_DEVICE_LEN 512
#define MAX_ERROR_LEN 512

#define FRAME_ADDRESS(mins, secs, frames) \
  (((mins) * FRAMES_PER_SECOND * 60) + ((secs) * FRAMES_PER_SECOND) + (frames))

static int debug;

static void yield_player(struct HJcd_Drive *drive);
static void take_player(struct HJcd_Drive *drive);

static void take_player(struct HJcd_Drive *drive)
{
  /*
   * If not already open, open the device.
   */
  if (unhand(drive)->fd == -1) {
    char device_name[MAX_DEVICE_LEN + 1];
    javaString2CString(unhand(drive)->device_name,device_name,MAX_DEVICE_LEN);
    if (device_name == '\0') {
      char *device = getenv("CDROM");
      if (device == NULL) {
	strcpy(device_name, "/dev/cdrom");
      }
      unhand(drive)->device_flags |= (getenv("SBPCD") != NULL) ?
	Jcd_Drive_FLAG_STOP_PLAY : 
	Jcd_Drive_FLAG_NONE; 
    }
    unhand(drive)->fd = open(device_name, O_RDONLY);

    if (debug)
      fprintf(stderr, "Openned %s flags=%d fd=%d\n ", 
	      device_name, 
	      unhand(drive)->device_flags,
	      unhand(drive)->fd);
  }
}


static void yield_player(struct HJcd_Drive *drive)
{
  /*
   * Close the device, making it avialable to others.
   */
  if (unhand(drive)->fd != -1) {
    close(unhand(drive)->fd);
  }

  unhand(drive)->fd               = -1;
  unhand(drive)->current_track    = 1;
  unhand(drive)->current_index    = 1;
  unhand(drive)->number_of_tracks = 0;
  unhand(drive)->audio_status     = Jcd_Drive_STATUS_INVALID;

  if (debug)
    fprintf(stderr, "closed cd device\n");
}

static Java_Int new_status(struct HJcd_Drive *drive)
{
  /* 
   * Try to obtain the device and query its status.
   * Could cause the tray to close with some drivers.
   */
  struct cdrom_subchnl ch;
  long stat;

  unhand(drive)->current_track   = 1;
  unhand(drive)->current_index   = 1;
  unhand(drive)->current_address = 0;


  if (unhand(drive)->fd == -1) { /* See if there's a new CD */
    take_player(drive);
    if (unhand(drive)->fd != -1) {
      struct cdrom_tochdr tinfo;
      stat = ioctl(unhand(drive)->fd, CDROMREADTOCHDR, &tinfo);
      if (stat == -1) {
	yield_player(drive);
	unhand(drive)->audio_status = Jcd_Drive_STATUS_INVALID;
	return (Java_Int) unhand(drive)->audio_status;
      }
      unhand(drive)->number_of_tracks = tinfo.cdth_trk1;
      if (debug)
	fprintf(stderr, "number_of_tracks=%d\n", unhand(drive)->number_of_tracks);
    }
  }

  ch.cdsc_format = CDROM_MSF;
    
  stat = ioctl(unhand(drive)->fd, CDROMSUBCHNL, &ch);
  if (stat == -1) {		/* Assume no CD in drive */
    yield_player(drive);
    unhand(drive)->audio_status = Jcd_Drive_STATUS_INVALID;
    return (Java_Int) unhand(drive)->audio_status;
  }

  unhand(drive)->current_track = ch.cdsc_trk;
  unhand(drive)->current_index = ch.cdsc_ind;

  unhand(drive)->current_address = FRAME_ADDRESS(ch.cdsc_absaddr.msf.minute,
						 ch.cdsc_absaddr.msf.second,
						 ch.cdsc_absaddr.msf.frame);

  unhand(drive)->audio_status = ch.cdsc_audiostatus;

  /* fprintf(stderr,"status=0x%x\n", audio_status); */
  return (Java_Int) unhand(drive)->audio_status;
}

static int is_open(struct HJcd_Drive *drive)
{
  if (unhand(drive)->fd == -1) {
    /* SignalError(0, "Jcd/DriveException", "Drive device not open."); */
    return FALSE;
  }
  return TRUE;
}

static int is_available_now(struct HJcd_Drive *drive)
{
  new_status(drive);  
  return is_open(drive);
}

void Jcd_Drive_initDrive(struct HJcd_Drive *drive)
{
  fprintf(stderr, "Initializing cd drive...\n");
  debug = getenv("JCD_DEBUG") != NULL;
  new_status(drive);
}

Java_Int Jcd_Drive_status(struct HJcd_Drive *drive)
{
  if (unhand(drive)->fd == -1) {
    return unhand(drive)->audio_status;
  }
  else {
    return new_status(drive);
  }
}

Java_Int Jcd_Drive_currentTrack(struct HJcd_Drive *drive) 
{
  if (!is_available_now(drive)) { return 1; }
  return unhand(drive)->current_track;
}

Java_Int Jcd_Drive_currentIndex(struct HJcd_Drive *drive) 
{				/* Must call current track first */
  if (!is_available_now(drive)) { return 1; }
  return unhand(drive)->current_index;
}


Java_Int Jcd_Drive_currentAddress(struct HJcd_Drive *drive) 
{
  if (!is_available_now(drive)) { return 0; }
  return unhand(drive)->current_address;
}


Java_Int Jcd_Drive_numberOfTracks(struct HJcd_Drive *drive)
{
  if (!is_open(drive)) { return 0; }
  return unhand(drive)->number_of_tracks;
}


Java_Int Jcd_Drive_trackAddress(struct HJcd_Drive *drive, Java_Int track)
{
  struct cdrom_tocentry tocentry;

  if (!is_open(drive)) { return 0; }
  tocentry.cdte_track = track;
  tocentry.cdte_format = CDROM_MSF;
  if ((ioctl(unhand(drive)->fd, CDROMREADTOCENTRY, &tocentry)) == -1) {
    if (debug)
      fprintf(stderr, "tae=%d\n", track);
    SignalError(0, "Jcd/TrackAddressException", strerror(errno));
    return 0;
  }
  return FRAME_ADDRESS(tocentry.cdte_addr.msf.minute,
		       tocentry.cdte_addr.msf.second,
		       tocentry.cdte_addr.msf.frame);
}

Java_Int Jcd_Drive_trackLength(struct HJcd_Drive *drive, Java_Int n)
{
  int starts_at = Jcd_Drive_trackAddress(drive, n);
  int start_of_next =
    (n >= unhand(drive)->number_of_tracks) 
    ? Jcd_Drive_cdEndAddress(drive) 
    : Jcd_Drive_trackAddress(drive, n + 1);
  return start_of_next - starts_at;
}


Java_Int Jcd_Drive_cdEndAddress(struct HJcd_Drive *drive)
{
  if (!is_open(drive)) { return 0; }
  return Jcd_Drive_trackAddress(drive, CDROM_LEADOUT);
}

static int cddbSum(int n) 
{				/* Sum the digits */
  int s = 0;
  while (n != 0) {
    s += n % 10;
    n = n / 10; 
  }
  return s;
}

struct Hjava_lang_String *Jcd_Drive_cddbID(struct HJcd_Drive *drive)
{
  /* xjcd cddb address - see http://sunsite.unc.edu~/cddb/xjcd/ */
  char id[10] = "00000000";
  int i;
  int t = 0;
  int n = 0;

  if (!is_open(drive)) { return makeJavaString(id, 8); }
  for (i = 1; i <= unhand(drive)->number_of_tracks; i++) {
    n += cddbSum(Jcd_Drive_trackAddress(drive, i) / FRAMES_PER_SECOND);
    t += (Jcd_Drive_trackLength(drive, i) / FRAMES_PER_SECOND) ;
  }
  i = ((n % 0xff) << 24 | t << 8 | (unhand(drive)->number_of_tracks));  
  sprintf(id, "%08x", i);
  return makeJavaString(id, 8);
}

void Jcd_Drive_play(struct HJcd_Drive *drive, 
			Java_Int start_track, 
			Java_Int start_index,
			Java_Int end_track,
			Java_Int end_index)
{
  struct cdrom_ti ti;
  long stat;

  unhand(drive)->current_track = 1;

  if (!is_available_now(drive)) { return; }

  if (debug)
    fprintf(stderr, "play %d %d %d\n", start_track, end_track, unhand(drive)->number_of_tracks);

  if (end_track == 0) {
    end_track = unhand(drive)->number_of_tracks; /* Kludge for first call to play. */
  }

  if (start_track < 1 || start_track > unhand(drive)->number_of_tracks) {
    SignalError(0, "Jcd/DriveException", "Play: start track out of range.");
    return;
  }

  if (end_track < start_track || end_track > unhand(drive)->number_of_tracks) {
    SignalError(0, "Jcd/DriveException", "Play: end track out of range.");
    return;
  }

  if (!is_open(drive)) {
    return;
  }
    
  if ((unhand(drive)->device_flags & Jcd_Drive_FLAG_STOP_PLAY)) {
    Jcd_Drive_stop(drive);	/* Must issue stop before play. */
  }

  if (debug)
    fprintf(stderr, "going to play %d\n", start_track);

  ti.cdti_trk0 = start_track;
  ti.cdti_ind0 = start_index;
  ti.cdti_trk1 = end_track;
  ti.cdti_ind1 = end_index;

  if (ti.cdti_ind1 == 0) {
    /* Doesn't seem to be a way of specifying end index.  But this
     * seems to work.  
     */
    ti.cdti_ind1 = unhand(drive)->number_of_tracks;
  }

  stat = ioctl(unhand(drive)->fd, CDROMPLAYTRKIND, &ti);
  if (stat < 0) {
    SignalError(0, "Jcd/PlayException", strerror(errno));
    return;
  }

  unhand(drive)->current_track = start_track;

  if (debug)
    fprintf(stderr, "playing %d\n", unhand(drive)->current_track);
}


void Jcd_Drive_stop(struct HJcd_Drive *drive)
{
  if (!is_available_now(drive)) { return; }
    
  if (ioctl(unhand(drive)->fd, CDROMSTOP) == -1) {
    SignalError(0, "Jcd/StopException", strerror(errno));
  }
}


void Jcd_Drive_pause(struct HJcd_Drive *drive)
{
  if (!is_available_now(drive)) { return; }
    
  if (unhand(drive)->audio_status != Jcd_Drive_STATUS_PLAY) {
    SignalError(0, "Jcd/PauseException", "Pause: drive isn't playing.");
    return;
  }
    
  if (ioctl(unhand(drive)->fd, CDROMPAUSE) == -1) {
    SignalError(0, "Jcd/PauseException", strerror(errno));
  }
}


void Jcd_Drive_resume(struct HJcd_Drive *drive)
{
  if (!is_available_now(drive)) { return; }
    
  if (unhand(drive)->audio_status != Jcd_Drive_STATUS_PAUSED) {
    SignalError(0, "Jcd/ResumeException", "Resume: drive isn't paused.");
    return;
  }

  if (ioctl(unhand(drive)->fd, CDROMRESUME) == -1) {
    SignalError(0, "Jcd/ResumeException", strerror(errno));
  }
}


void Jcd_Drive_eject(struct HJcd_Drive *drive)
{
  long retract = unhand(drive)->fd == -1;

  if (!is_available_now(drive)) { return; }

  if (retract) {
    return;
  }
    
  Jcd_Drive_stop(drive);
    
  if (ioctl(unhand(drive)->fd, CDROMEJECT) == -1) {
    SignalError(0, "Jcd/EjectException", strerror(errno));
  }

  yield_player(drive);
}


Java_Int Jcd_Drive_volume(struct HJcd_Drive *drive)
{
  struct cdrom_volctrl vol;

  if (!is_available_now(drive)) { return; }

  if (ioctl(unhand(drive)->fd, CDROMVOLREAD, &vol) == -1) {
    SignalError(0, "Jcd/VolumeException", strerror(errno));
    return 0;
  }
  return vol.channel0;
}

void Jcd_Drive_setVolume(struct HJcd_Drive *drive, Java_Int volume)
{
  struct cdrom_volctrl vol;

  if (!is_available_now(drive)) { return; }

  if (volume < 0 || volume > 255) {
    SignalError(0, "Jcd/SetVolumeException", "Volume out of range.");
    return;
  }

  vol.channel0 = volume;
  vol.channel1 = volume;
  vol.channel2 = volume;
  vol.channel3 = volume;

  if (ioctl(unhand(drive)->fd, CDROMVOLCTRL, &vol) == -1) {
    SignalError(0, "Jcd/SetVolumeException", strerror(errno));
  }
}


struct Hjava_lang_String * Jcd_Drive_productCode(struct HJcd_Drive *drive)
{
  char upc[9] = "00000000";

  if (!is_available_now(drive)) { return makeJavaString(upc, 8); }
    
  if (ioctl(unhand(drive)->fd, CDROM_GET_UPC, upc) == -1) {
    SignalError(0, "Jcd/ProductCodeException", strerror(errno));
  }
  return makeJavaString(upc, 8);
}

