package Jcd;

/**
  Jcd - Java CD Audio Player
  Copyright (c) 1996.   Michael Hamilton (michael@actrix.gen.nz).
  All rights reserved.  See the README for full details
 */

import java.io.*;
import java.util.*;
import java.awt.*;
import Jcd.Drive;
import Jcd.Monitor;
import Jcd.JcdProps;

class SmartDrive extends Drive implements Observer {

  static final boolean DEBUG = JcdProps.getBoolean("jcd.debug", false); 

  public Monitor monitor;

  public TrackList tracksToPlay = null;
  public int program_pos = 0;
  public boolean singleMode = false;

  protected boolean singleModePending = false;


  public SmartDrive(String device, String module, int flags) 
  {
    super(device, module, flags);
    setUp();
  }

  protected void setUp()
  {
    monitor = new Monitor(this);
    monitor.addObserver(this);
    tracksToPlay = new TrackList(30);
  }

  public void setSingleMode(boolean state)
  {
    singleMode = state;

    // If playing, set a flag so we can switch play modes at the end of the
    // track.
    singleModePending = monitor.status == Drive.STATUS_PLAY;
  }

  public void play(int track) throws DriveException
  {
    if (singleMode) {
      single(track);
    } 
    else {
      super.play(track);
    }
  }

  public void eject() throws DriveException
  {
    super.eject();
  }

  public boolean startPlaying() throws DriveException
  {
    if (status() == Drive.STATUS_PAUSED) {
      resume();
    }
    else if (!tracksToPlay.isEmpty()) {
      if (tracksToPlay.atEnd()) {
	tracksToPlay.reset();
      }
      play(tracksToPlay.nextTrack());
    }
    else {
      play(1);
    }
    return true;
  }

  public boolean next() throws DriveException
  {
    if (tracksToPlay.isEmpty()) {
      int track = currentTrack();
      int max = numberOfTracks();
      if (track >= max) {
	return false;
      }
      play(track + 1);
    }
    else {
      int next = tracksToPlay.nextTrack();
      if (next <= 0) {
	return false;
      }
      play(next);
    }
    return true;
  }

  public boolean prev() throws DriveException
  {
    if (tracksToPlay.isEmpty()) {
      int track = currentTrack();
      if (track <= 1) {
	return false;
      }
      play(track - 1);
    }
    else {
      int next = tracksToPlay.prevTrack();
      if (next <= 0) {
	return false;
      }
      play(next);
    }
    return true;
  }

  public int remaining() 
  {				// Should this method belong to Monitor?
    int result = 0;
    if (tracksToPlay.isEmpty()) {
      result = monitor.cdEndAddress - monitor.currentAddress;
    }
    else {			// Should really do some caching here.
				// But no serious impact on CPU, so lets be
				// lazy for now.
      int index = 0;
      Enumeration iter = tracksToPlay.elements();
      while (iter.hasMoreElements()) {
	int track = ((Integer) iter.nextElement()).intValue();
	if (index >= tracksToPlay.index 
	    && track < monitor.trackLength.length) {
	  result += monitor.trackLength[track];
	}
	index++;
      }
      if (!tracksToPlay.atStart() && !tracksToPlay.atEnd()) {
	result -= (monitor.currentAddress - monitor.trackStartAddress);
      }
    }
    return result;
  }

  public int togglePause() throws DriveException
  {
    if (status() == Drive.STATUS_PAUSED) {
      resume();
      return Drive.STATUS_PLAY;
    }
    pause();
    return Drive.STATUS_PAUSED;
  }

  public synchronized void update(Observable o, Object arg)
  {
    try {
      if (monitor.status != Drive.STATUS_INVALID) {
	if (monitor.cdChanged) {

	  if (DEBUG) {
	    System.out.println("New CD");
	    System.out.println("Id=" + productCode());
	  }
	  tracksToPlay.clear();
	}
	else if (monitor.status == Drive.STATUS_PLAY) {

	  int track = monitor.currentTrack;
	  int tend =  monitor.trackEndAddress[track];

	  if (monitor.currentAddress >= tend - 210) { // Close to end of track?

	    // Poll frequently so we don't miss the event. 
	    while (currentAddress() < tend) {
	      try {
		Thread.sleep(100);	   // Sleep 100 msec's.
	      }
	      catch (InterruptedException e) {
	      }
	    }

	    if (!tracksToPlay.atStart()) { // Programme play
	      programEndOfTrack();
	    }
	    else {			   // Normal play
	      normalEndOfTrack(track);
	    }

	    singleModePending = false;	   // Finished end of track processing.
	  }
	}
      }
    }
    catch (DriveException e) {
      System.out.println("Exception: " + e);
    }
  }

  protected void programEndOfTrack()
  {
    try {
      int next = tracksToPlay.nextTrack(); 
      if (tracksToPlay.atEnd()) {
	if (DEBUG) System.out.println("end");
	stop();
	tracksToPlay.reset();
      }
      else if (singleMode) {
	stop();
      }
      else {
	if (DEBUG) System.out.println("next");
	play(next);
      }
    }
    catch (DriveException e) {
      System.out.println("Program end of track exception: " + e);
    }
  }
  
  protected void normalEndOfTrack(int track)
  {					   
    // If during the track we switched play modes, ensure the correct
    // behaviour occurs.

    if (singleModePending) {
      try {
	if (singleMode) {
	  stop();
	}
	else if (track < monitor.numberOfTracks) {
	  play(track + 1);
	}
      }
      catch (DriveException e) {
	System.out.println("Normal-Play end of track exception: " + e);
      }      
    }
  }
}


class TrackList extends Vector {

  static final boolean DEBUG = JcdProps.getBoolean("jcd.debug", false); 

  public int index = -1;

  protected Integer trackId[];	// Unique Object for each track

  protected boolean at_end = false;

  TrackList(int n)
  {
    trackId = new Integer[n];
    for (int i = 0; i < n; i++) {
      trackId[i] = new Integer(i + 1);
    }
  }

  public void addTrack(int t) 
  {
    addElement(trackId[t - 1]);
  }

  public void removeTrack(int t)
  {
    int pos = lastIndexOf(trackId[t - 1]);
    if (pos >= 0) {
      removeElement(trackId[t - 1]);
    }
  }
  
  public int currentTrack() 
  {
    if (index < 0) {
      return 0;
    }
    return ((Integer) elementAt(index)).intValue();
  }

  public int nextTrack()
  {
    index++;
    if (index >= size()) {
      index = size() - 1;
      at_end = true;
    }
    if (DEBUG) System.out.println("next track =" + index);
    return ((Integer) elementAt(index)).intValue();
  }

  public int prevTrack()
  {
    index--;
    if (index <= 0) {
      index = 0;
    }
    return ((Integer) elementAt(index)).intValue();
  }

  public boolean atStart()
  {				// -1 == Haven't started playing the list yet.
    return index == -1;
  }

  public boolean atEnd()
  {
    return at_end;
  }

  public void clear() 
  {
    index = -1;
    at_end = false;
    removeAllElements();
  }

  public void reset() 
  {
    at_end = false;
    index = -1;
  }

}
